"""faligning comments and variable declarations."""
from typing import List, Optional


class FAlign:
    """Aligning comments and sometimes variable declarations. """

    def __init__(self, min_space_before_comment: int = 1,
                 comment_start_column: Optional[int] = None):
        self.min_space_before_comment = min_space_before_comment
        self.comment_start_column = comment_start_column

    def falign(self, lines: List[str]) -> List[str]:
        """aligns input file and writes to output file."""

        output_lines: List[str] = []
        block_slice: slice = slice(None)
        while True:
            block_slice = self.next_block_slice(block_slice.stop, lines)
            if block_slice.start is None:
                break

            # Determine the code indent at the start of the next block for
            # trailing comments and comments outside a block
            next_block_code_indent: Optional[int] = None
            next_block_slice: slice = block_slice
            while next_block_code_indent is None:
                next_block_slice = \
                    self.next_block_slice(next_block_slice.stop, lines)
                if next_block_slice.start is None:
                    break
                else:
                    next_block_code_indent = \
                        FAlign._code_indent(lines[next_block_slice])

            block = lines[block_slice]
            self._align_block_comments(block, next_block_code_indent)
            output_lines.extend(block)

        return output_lines

    def _align_block_comments(self, lines: List[str],
                              next_block_code_indent: Optional[int] = None
                              ) -> None:
        to_align: List[int] = []
        align_index: int = 0 if self.comment_start_column is None \
            or self.comment_start_column <= 0 \
            else self.comment_start_column - 1

        for iline, line in enumerate(lines):
            comment_index: Optional[int] = self._comment_index(line)

            # Leave comments without indentation at the beginning
            if comment_index is None or comment_index == 0:
                continue

            code_length = len(line[:comment_index].rstrip())

            if code_length + self.min_space_before_comment > align_index:
                align_index = code_length + self.min_space_before_comment

            # Comment only line
            if self._first_non_whitespace(lines[iline]) == '!':
                code_indent: Optional[int] = self._code_indent(lines[iline:])

                # If we could not determine the code indent of the next line
                # of code use the one found in the next block
                if code_indent is None:
                    code_indent = next_block_code_indent

                # Fix comments which belong to code and are not indented enough
                if code_indent is not None and 0 < comment_index < code_indent:
                    lines[iline] = ' ' * code_indent + \
                                   lines[iline][comment_index:]
                    continue
                # Ignore comments which belong to code and are on
                # code indention
                elif comment_index == code_indent:
                    continue
                # Align comments whose indent is greater than
                # current code indent
                else:
                    pass

            to_align.append(iline)

        for iline in to_align:
            comment_index = self._comment_index(lines[iline])

            if comment_index is None or comment_index == 0:
                continue

            lines[iline] = lines[iline][:comment_index-1] \
                + ' ' * (align_index + 1 - comment_index) \
                + lines[iline][comment_index:]

    @staticmethod
    def _code_indent(lines: List[str]) -> Optional[int]:
        """Gives the current indent of the current line of code, if the current
        line contains no code looks for the next line of code."""

        for iline, _ in enumerate(lines):
            if FAlign._is_empty(lines[iline]):
                continue
            inws: Optional[int] = FAlign._first_non_whitespace_index(
                lines[iline])
            if inws is None:
                continue
            if lines[iline][inws] == '!':
                continue
            return inws
        return None

    @staticmethod
    def _comment_index(line: str) -> Optional[int]:
        string_delimiter: Optional[str] = None

        for iline, _ in enumerate(line):
            char = line[iline]

            if string_delimiter and char != string_delimiter:
                continue

            if string_delimiter and char == string_delimiter:
                string_delimiter = None
                continue

            if char in ('\'', '"'):
                string_delimiter = char
                continue

            if char == '!':
                return iline

        return None

    @staticmethod
    def _first_non_whitespace(line: str) -> Optional[str]:
        i = FAlign._first_non_whitespace_index(line)
        return None if i is None else line[i]

    @staticmethod
    def _first_non_whitespace_index(line: str) -> Optional[int]:
        for iline, _ in enumerate(line):
            if line[iline].strip():
                return iline

        return None

    @staticmethod
    def next_block_slice(current: Optional[int], lines: List[str]) \
            -> slice:
        """Finds next block of code and includes leading and trailing empty
        lines."""

        if not current:
            start = 0
        elif current < 0:
            start = 0
        else:
            start = current

        if current is not None and current >= len(lines):
            return slice(None)

        end = start

        # Find end of block
        while True:
            # End of list
            if end >= len(lines):
                return slice(start, end)

            # Empty line --> search for start of next block --> end of block
            if FAlign._is_empty(lines[end]):
                # Find first non empty line and include trailing empty lines
                while True:
                    # End of list
                    if end >= len(lines):
                        return slice(start, end)

                    # Non empty line --> End of block
                    if not FAlign._is_empty(lines[end]):
                        return slice(start, end)

                    end = end + 1

            end = end + 1

    @staticmethod
    def _is_empty(string: str) -> bool:
        return not string.strip()



import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="fortran-align",
    version="0.0.1",
    author="Tobias R. Henle",
    author_email="tobias@page23.de",
    description="Fortran comment alignment tool",
    entry_points={'console_scripts': ['falign = falign.__main__:run']},
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/TobiasRH/fortran-align",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7.2',
)



"""Align comments in Fortran files."""
import argparse
import sys
import logging
from typing import List
from falign.falign import FAlign


def run():
    """Run FAlign."""
    parser = argparse.ArgumentParser(prog='falign')
    parser.add_argument('--output', default='',
                        metavar='file-name', help='Output file')
    parser.add_argument('--inplace', default=False,
                        action='store_true', help='Reformat inplace')
    parser.add_argument('--dry-run', default=False,
                        action='store_true', help='Just write result to standard error and do not touch output file')
    parser.add_argument('--debug', default=False,
                        action='store_true', help='Write debugging information to standard error')
    parser.add_argument('input', nargs='?', default='',
                        metavar='file-name', help='Input file')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    else:
        logging.basicConfig(stream=sys.stderr, level=logging.WARNING)

    if args.inplace and args.output:
        logging.error(
            'Error in Argument: --inplace and --output can not be defined at the same time!')
        sys.exit(1)

    if args.inplace and not args.input:
        logging.error(
            'Error in Argument: --inplace can not be used without input file!')
        sys.exit(1)

    input_lines: List[str] = []
    output_lines: List[str] = []

    if args.input:
        logging.debug('open(%s)', args.input)
        input_file = open(args.input, mode='r', encoding='ascii',
                          errors='surrogateescape')
        logging.debug('read(%s)', args.input)
        input_lines.extend(input_file.readlines())
        logging.debug('close(%s)', args.input)
        input_file.close()
    else:
        logging.debug('read(sys.stdin)')
        input_lines.extend(sys.stdin.readlines())

    if args.debug:
        sys.stderr.write(''.join(input_lines))

    logging.debug('falign()')
    falign: FAlign = FAlign(1, 49)
    output_lines.extend(falign.falign(input_lines))

    if args.inplace and input_lines == output_lines:
        logging.debug('input equals output: no change --> no write.')
    elif args.dry_run:
        logging.debug('dry-run: write(stderr)')
        sys.stderr.write('\n'.join(output_lines))
    elif args.output:
        logging.debug('open(%s)', args.output)
        output_file = open(args.output, mode='w', encoding='ascii',
                           errors='surrogateescape')
        logging.debug('write(%s)', args.output)
        output_file.writelines(output_lines)
        logging.debug('%s.close()', args.output)
        output_file.close()
    elif args.inplace and args.input:
        logging.debug('open(%s)', args.input)
        output_file = open(args.input, mode='w', encoding='ascii',
                           errors='surrogateescape')
        logging.debug('write(%s)', args.input)
        output_file.writelines(output_lines)
        logging.debug('%s.close()', args.input)
        output_file.close()
    else:
        logging.debug('write(stdout)')
        sys.stdout.writelines(output_lines)

    logging.debug('done.')


if __name__ == '__main__':
    run()
